import sys
import numpy as np

from warnings import catch_warnings, simplefilter
from argparse import Namespace
from rich.console import Console
from rich.progress import Progress
from MDAnalysis import Universe, AtomGroup
from MDAnalysis.coordinates.base import Timestep, ReaderBase
from sklearn.cluster import DBSCAN

from .cli_args import parse_cli_args
from .ndx_file import read_ndx_file

console = Console()


def get_ref_selection(universe: Universe, args: Namespace) -> AtomGroup:

    ref_selection: AtomGroup
    if args.mode == "CA":
        ref_selection = universe.select_atoms("name CA")
    elif args.mode == "CB":
        ref_selection = universe.select_atoms("name CB or (name CA and resname GLY)")
    elif args.mode == "index":
        ref_selection = read_ndx_file(args.index_file, universe, console)
    else:
        raise Exception(f"Mode error! Mode was \"{args.mode}\", which is unexpected!")

    console.print(f"Number of selected residues: {ref_selection.n_residues}")
    console.print(f"Number of selected atoms: {ref_selection.n_atoms}")

    return ref_selection


def get_trajectory_slice(universe: Universe, args: Namespace) -> ReaderBase:

    begin_frame = int(args.begin_time / universe.trajectory.dt)

    console.print(f"Start frame is {begin_frame} at time {args.begin_time} ps.")

    if args.end_time is None:
        end_frame: int = len(universe.trajectory)
        args.end_time = end_frame * universe.trajectory.dt
    else:
        end_frame = int(args.end_time / universe.trajectory.dt)

    console.print(f"End frame is {end_frame} at time {args.end_time} ps.")

    if args.delta_time is None:
        delta_frame: int = 1
        args.delta_time = universe.trajectory.dt
    else:
        delta_frame = int(args.delta_time / universe.trajectory.dt)

    if delta_frame == 0:
        console.print(
            f"[red]ERROR: "
            f"The delta time for the analysis ({args.delta_time} ps) "
            f"is smaller than the delta time of the trajectory ({universe.trajectory.dt} ps)!"
        )
        sys.exit(1)

    console.print(f"Delta frame is {delta_frame} with a delta time of {args.delta_time} ps.")

    trajectory = universe.trajectory[begin_frame:end_frame:delta_frame]

    return trajectory


def get_distance_std_mx(
    trajectory: ReaderBase,
    ref_selection: AtomGroup,
    args: Namespace
) -> np.ndarray:

    dmx_mean = None
    dmx_variance = None

    with Progress(console=console) as progress:

        task = progress.add_task("Calculating matrices...", total=len(trajectory))

        frame: Timestep
        for frame_idx, frame in enumerate(trajectory):

            ca_positions: np.ndarray = ref_selection.positions
            dmx = ca_positions[:, np.newaxis, :] - ca_positions[np.newaxis, :, :]
            dmx = np.sqrt(np.sum(dmx ** 2, axis=2))

            if dmx_mean is None:
                dmx_mean = np.copy(dmx)
                dmx_variance = np.zeros(dmx.shape)
            else:
                new_dmx_mean = dmx_mean + (dmx - dmx_mean) / (frame_idx + 1)
                dmx_variance += (dmx - dmx_mean) * (dmx - new_dmx_mean)
                dmx_mean = np.copy(new_dmx_mean)

            current_time = args.begin_time + frame_idx * args.delta_time
            progress.update(task, advance=1)  # , description=f"Sim. time: {current_time} ps")

    return np.sqrt(dmx_variance / len(trajectory))


def show_dist_std_mx_stat(distance_std_mx: np.ndarray, args: Namespace) -> None:

    console.print(f"Distance standard deviation matrix constructed with shape: {distance_std_mx.shape}.")

    console.print(f"\tMin: {np.min(distance_std_mx):.3f} nm")
    console.print(f"\tMax: {np.max(distance_std_mx):.3f} nm")
    console.print(f"\tAvg: {np.mean(distance_std_mx):.3f} nm")
    console.print(f"\tMedian: {np.median(distance_std_mx):.3f} nm")
    console.print(f"\tStd: {np.std(distance_std_mx):.3f} nm")

    nth_closest_neigh_med_value = np.median(np.sort(distance_std_mx, axis=1)[:, args.cluster_neighbours])
    console.print(f"\tMedian of {args.cluster_neighbours}th closest neighbour: {nth_closest_neigh_med_value:.3f} nm")

    if args.cluster_radius is None:
        console.print("[yellow]No clustering radius was given! Using the previous median value as a radius.")
        args.cluster_radius = nth_closest_neigh_med_value


def perform_segmentation(distance_std_mx: np.ndarray, args: Namespace) -> np.ndarray:

    segmentation_labels: np.ndarray = DBSCAN(
        eps=args.cluster_radius,
        min_samples=args.cluster_neighbours,
        metric="precomputed"
    ).fit(distance_std_mx).labels_

    console.print(f"Segmentation done! Found {len(np.unique(segmentation_labels))} segments.")

    for idx in np.sort(np.unique(segmentation_labels)):
        current_size = np.sum(segmentation_labels == idx)
        console.print(f"\tSize of segment {idx}: {current_size}")

    return segmentation_labels


def blabel(
    universe: Universe,
    ref_selection: AtomGroup,
    segmentation_labels: np.ndarray,
    args: Namespace
):

    console.print("Resetting atom bfactors...")
    universe.add_TopologyAttr("tempfactors")  # needed to set the tempfactors of the atoms
    for atom in universe.atoms:
        atom.tempfactor = -1.

    if args.mode == "index":
        for idx, atom in enumerate(ref_selection.atoms):
            atom.tempfactor = segmentation_labels[idx]
    else:
        for idx, atom in enumerate(ref_selection.atoms):
            tempfactor_array = segmentation_labels[idx] * np.ones(len(atom.residue.atoms))
            atom.residue.atoms.tempfactors = tempfactor_array

    # Without "catch_warnings" we would get "Found no information for attr: 'formalcharges'".
    with catch_warnings():
        simplefilter(action="ignore")
        universe.select_atoms("all").write(args.output_file)

    console.print(f"B-labelled pdb file saved as {args.output_file}")


def main():

    console.print("[green](main) Parsing CLI arguments...")
    args = parse_cli_args(console)

    console.print("[green](main) Reading trajectory and structure files...")
    universe = Universe(args.structure_file, args.trajectory_file)

    console.print("[green](main) Getting the reference selection...")
    ref_selection = get_ref_selection(universe, args)

    console.print("[green](main) Getting the analyzable trajectory interval...")
    trajectory = get_trajectory_slice(universe, args)

    console.print("[green](main) Calculating the distance standard deviation matrix...")
    distance_std_mx = get_distance_std_mx(trajectory, ref_selection, args)

    console.print("[green](main) Displaying statistics about the distance standard deviation matrix...")
    show_dist_std_mx_stat(distance_std_mx, args)

    console.print("[green](main) Performing segmentation...")
    segmentation_labels = perform_segmentation(distance_std_mx, args)

    console.print("[green](main) Saving the b-labelled structure...")
    blabel(universe, ref_selection, segmentation_labels, args)

    console.print("[green](main) Program terminated with success!")


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

setup(
    name="rigid_body_segmentation",
    version="0.0.2",
    packages=find_packages(),
    install_requires=[
        "MDAnalysis>=2.7.0",
        "numpy>=1.26.3",
        "scikit-learn>=1.4.0",
        "rich>=13.7.0"
    ]
)



import sys
import numpy as np

from rich.console import Console
from collections import OrderedDict
from MDAnalysis import AtomGroup, Universe


def read_ndx_file(ndx_file_path: str, universe: Universe, console: Console) -> AtomGroup:

    ndx_content = OrderedDict()

    # Reading the ndx file
    with open(ndx_file_path, "r") as f:
        ndx_file = f.read()
    ndx_file = ndx_file.split("\n")

    # Parsing the ndx file content
    current_key = None
    for line in ndx_file:

        if line.startswith("[ "):
            current_key = line[2:-2]
            ndx_content[current_key] = list()
        elif line.startswith(";"):
            pass
        else:
            ndx_content[current_key] += list(map(int, line.split()))

    # Printing the groups in GROMACS style
    max_name_length = max(map(len, ndx_content.keys()))
    max_len_length = max(map(lambda x: len(str(len(x))), ndx_content.values()))

    console.print("Select group for rigid body analysis")
    for group_idx, group_name in enumerate(ndx_content):

        str_len = str(len(ndx_content[group_name]))

        current_line = "Group"
        current_line += str(group_idx).rjust(6)
        current_line += f" ( {group_name.ljust(max_name_length)} ) has "
        current_line += f"{str_len.ljust(max_len_length)} elements"

        console.print(current_line)

    selected_idx = int(console.input("Select a group: "))

    # Getting the selected atom IDs
    ndx_content = list(ndx_content.values())
    if selected_idx >= len(ndx_content) or selected_idx < 0:
        console.print(
            f"[red]ERROR: Invalid group index: {selected_idx}!"
        )
        sys.exit(1)

    atom_ids = ndx_content[selected_idx]

    # Getting the selected atoms.
    # This is highly inefficient, but will do for now...
    atom_indices = list()
    for current_id in atom_ids:

        current_idx = np.argwhere(universe.atoms.ids == current_id)
        if len(current_idx) != 1:
            console.print(
                f"[red]ERROR: Atom ID not found: {current_id}! "
                f"Maybe the given structure does not have atoms that are specified in the selected group?"
            )
            sys.exit(1)
        current_idx = current_idx[0, 0]
        atom_indices.append(current_idx)

    atom_indices = np.array(atom_indices)
    atom_group = universe.atoms[atom_indices]

    return atom_group



import sys
from rich.console import Console
from argparse import ArgumentParser, Namespace
from pathlib import Path


def _validate_arguments(args: Namespace, console: Console):

    if not Path(args.trajectory_file).exists():
        console.print("[red]ERROR: Path to trajectory file does not exist!")
        sys.exit(1)

    if not Path(args.trajectory_file).is_file():
        console.print("[red]ERROR: Path to trajectory file is not a file!")
        sys.exit(1)

    if not args.trajectory_file.endswith(".xtc"):
        console.print("[red]ERROR: Trajectory file does not have an xtc extension!")
        sys.exit(1)

    if not Path(args.structure_file).exists():
        console.print("[red]ERROR: Path to structure file does not exist!")
        sys.exit(1)

    if not Path(args.structure_file).is_file():
        console.print("[red]ERROR: Path to structure file is not a file!")
        sys.exit(1)

    if not args.structure_file.endswith(".pdb"):
        console.print("[red]ERROR: Structure file does not have a pdb extension!")
        sys.exit(1)

    if args.index_file is not None and args.mode == "index":

        if not Path(args.index_file).exists():
            console.print("[red]ERROR: Path to index file does not exist!")
            sys.exit(1)

        if not Path(args.index_file).is_file():
            console.print("[red]ERROR: Path to index file is not a file!")
            sys.exit(1)

        if not args.index_file.endswith(".ndx"):
            console.print("[red]ERROR: Index file does not have an ndx extension!")
            sys.exit(1)

    if not args.output_file.endswith(".pdb"):
        args.output_file += ".pdb"


def parse_cli_args(console: Console) -> Namespace:

    arg_parser = ArgumentParser()

    arg_parser.add_argument(
        "-f", "--trajectory_file",
        type=str, required=True,
        help="The path to the trajectory file. It's extension should be xtc."
    )

    arg_parser.add_argument(
        "-s", "--structure_file",
        type=str, required=True,
        help="The path to the structure file. It's extension should be pdb."
    )

    arg_parser.add_argument(
        "-o", "--output_file",
        type=str, required=True,
        help="The b-labelled output pdb file."
    )

    arg_parser.add_argument(
        "-n", "--index_file",
        type=str, default=None,
        help="The index file containing the atoms for the RBS if mode is \"index\"."
    )

    arg_parser.add_argument(
        "-clr", "--cluster_radius",
        type=float, default=None,
        help="The epsilon value for the DBSCAN clustering algorithm given in nm."
    )

    arg_parser.add_argument(
        "-cln", "--cluster_neighbours",
        type=int, default=5,
        help="The min_samples value for the DBSCAN clustering algorithm."
    )

    arg_parser.add_argument(
        "-b", "--begin_time",
        type=int, default=0,
        help="The time (in ps) from the algorithm reads the trajectory."
    )

    arg_parser.add_argument(
        "-e", "--end_time",
        type=int, default=None,
        help="The time (in ps) until the algorithm reads the trajectory."
    )

    arg_parser.add_argument(
        "-dt", "--delta_time",
        type=int, default=None,
        help="The time (in ps) interval the algorithm reads the trajectory."
    )

    arg_parser.add_argument(
        "-m", "--mode",
        type=str, default="CA", choices=["CA", "CB", "index"],
        help="Either \"CA\", \"CB\", or \"index\"."
    )

    args: Namespace = arg_parser.parse_args()
    _validate_arguments(args, console)

    return args
